from __future__ import annotations

from typing import Any


def cosine_similarity_loop(query: Any, vectors: Any):
    """Cosine similarity baseline using a pure Python loop.

    Restrictions (NFR-1):
      - Use numpy.dot + numpy.linalg.norm inside the for-loop.
    """
    import numpy as np

    q = np.asarray(query, dtype=np.float32)
    v = np.asarray(vectors, dtype=np.float32)
    if v.ndim != 2:
        raise ValueError("vectors must be a 2D array")
    if q.ndim != 1:
        raise ValueError("query must be a 1D array")
    if v.shape[1] != q.shape[0]:
        raise ValueError("dimension mismatch")

    q_norm = float(np.linalg.norm(q))
    out = np.empty((v.shape[0],), dtype=np.float32)

    denom_q = q_norm if q_norm != 0.0 else 1.0
    for i in range(v.shape[0]):
        dot = float(np.dot(q, v[i]))
        v_norm = float(np.linalg.norm(v[i]))
        denom = denom_q * (v_norm if v_norm != 0.0 else 1.0)
        out[i] = dot / denom

    return out


def cosine_similarity_vectorized(query: Any, vectors: Any):
    """Reference cosine similarity via vectorized NumPy operations."""
    import numpy as np

    q = np.asarray(query, dtype=np.float32)
    v = np.asarray(vectors, dtype=np.float32)
    if v.ndim != 2:
        raise ValueError("vectors must be a 2D array")
    if q.ndim != 1:
        raise ValueError("query must be a 1D array")
    if v.shape[1] != q.shape[0]:
        raise ValueError("dimension mismatch")

    denom = np.linalg.norm(q) * np.linalg.norm(v, axis=1)
    denom = np.where(denom == 0.0, 1.0, denom)
    return np.dot(q, v.T) / denom


def cosine_similarity_loop_over_get_vector(txn: Any, keys: list[bytes], query: Any, metric: Any):
    """Cosine similarity baseline using a Python loop + per-key get_vector.

    This represents a naive Python approach:
      - iterate keys in Python
      - fetch each vector via the bindings (N calls)
      - compute cosine similarity with numpy.dot + numpy.linalg.norm in the loop
    """
    import numpy as np

    q = np.asarray(query, dtype=np.float32)
    if q.ndim != 1:
        raise ValueError("query must be a 1D array")

    q_norm = float(np.linalg.norm(q))
    denom_q = q_norm if q_norm != 0.0 else 1.0

    out = np.empty((len(keys),), dtype=np.float32)
    for i, key in enumerate(keys):
        v = txn.get_vector(key, metric)
        dot = float(np.dot(q, v))
        v_norm = float(np.linalg.norm(v))
        denom = denom_q * (v_norm if v_norm != 0.0 else 1.0)
        out[i] = dot / denom

    return out



import pytest

from alopex import AlopexError
from alopex._alopex import catalog as _catalog


def test_resolve_credentials_local_path_returns_empty(monkeypatch):
    monkeypatch.delenv("AWS_ACCESS_KEY_ID", raising=False)
    monkeypatch.delenv("AWS_SECRET_ACCESS_KEY", raising=False)
    result = _catalog._resolve_credentials("data/table")
    assert result == {}


def test_resolve_credentials_file_scheme_returns_empty(monkeypatch):
    monkeypatch.delenv("AWS_ACCESS_KEY_ID", raising=False)
    monkeypatch.delenv("AWS_SECRET_ACCESS_KEY", raising=False)
    result = _catalog._resolve_credentials("file:///tmp/data.parquet")
    assert result == {}


def test_resolve_credentials_s3_uses_env(monkeypatch):
    monkeypatch.setenv("AWS_ACCESS_KEY_ID", "dummy_access")
    monkeypatch.setenv("AWS_SECRET_ACCESS_KEY", "dummy_secret")
    monkeypatch.setenv("AWS_REGION", "ap-northeast-1")
    result = _catalog._resolve_credentials("s3://bucket/path")
    assert result["aws_access_key_id"] == "dummy_access"
    assert result["aws_secret_access_key"] == "dummy_secret"
    assert result["aws_region"] == "ap-northeast-1"


def test_resolve_credentials_gcs_uses_env(monkeypatch):
    monkeypatch.setenv("GOOGLE_APPLICATION_CREDENTIALS", "/tmp/gcs.json")
    result = _catalog._resolve_credentials("gs://bucket/path")
    assert result["service_account_path"] == "/tmp/gcs.json"


def test_resolve_credentials_azure_uses_env(monkeypatch):
    monkeypatch.setenv("AZURE_STORAGE_ACCOUNT", "account")
    monkeypatch.setenv("AZURE_STORAGE_KEY", "key")
    result = _catalog._resolve_credentials("az://container/path")
    assert result["account_name"] == "account"
    assert result["account_key"] == "key"


def test_resolve_credentials_unknown_scheme_returns_empty(monkeypatch):
    monkeypatch.delenv("AWS_ACCESS_KEY_ID", raising=False)
    monkeypatch.delenv("AWS_SECRET_ACCESS_KEY", raising=False)
    result = _catalog._resolve_credentials("ftp://example.com/data")
    assert result == {}


def test_resolve_credentials_dict_and_storage_options():
    result = _catalog._resolve_credentials(
        "s3://bucket/path",
        credential_provider={"token": "old", "shared": "old"},
        storage_options={"shared": "new", "extra": "1"},
    )
    assert result["token"] == "old"
    assert result["shared"] == "new"
    assert result["extra"] == "1"


def test_resolve_credentials_invalid_provider_raises():
    with pytest.raises(AlopexError):
        _catalog._resolve_credentials("s3://bucket/path", "unsupported")


def test_resolve_credentials_invalid_type_raises():
    with pytest.raises(AlopexError):
        _catalog._resolve_credentials("s3://bucket/path", ["not", "dict"])



from . import _alopex as _alopex
from ._alopex import catalog as _catalog  # type: ignore[attr-defined]
from ._alopex import database as _database  # type: ignore[attr-defined]
from ._alopex import transaction as _transaction  # type: ignore[attr-defined]
from ._alopex import types as _types  # type: ignore[attr-defined]

_EXCLUDE = {"catalog", "database", "transaction", "types"}
_seen = set()
__all__ = []


def _export_public(module) -> None:
    for _name in dir(module):
        if _name.startswith("_") or _name in _EXCLUDE or _name in _seen:
            continue
        globals()[_name] = getattr(module, _name)
        __all__.append(_name)
        _seen.add(_name)


_export_public(_alopex)
_export_public(_database)
_export_public(_transaction)
_export_public(_types)
_export_public(_catalog)

for _name in ("Catalog", "CatalogInfo", "NamespaceInfo", "TableInfo", "ColumnInfo"):
    globals()[_name] = getattr(_catalog, _name)
    if _name not in _seen:
        __all__.append(_name)
        _seen.add(_name)



import pytest

from alopex import AlopexError, Database, EmbeddedConfig, TxnMode


def test_open_in_memory():
    db = Database.open_in_memory()
    stats = db.memory_usage()
    assert stats.total_bytes >= 0
    db.close()


def test_new_is_in_memory(db):
    stats = db.memory_usage()
    assert stats.total_bytes >= 0


def test_open_with_config_in_memory():
    config = EmbeddedConfig()
    db = Database.open_with_config(config)
    db.close()


def test_begin_default_is_read_only(db):
    txn = db.begin()
    with pytest.raises(AlopexError):
        txn.put(b"key", b"value")


def test_begin_read_write_allows_put(db):
    txn = db.begin(TxnMode.READ_WRITE)
    txn.put(b"key", b"value")
    txn.commit()


def test_close_rolls_back_active_transaction(db):
    txn = db.begin(TxnMode.READ_WRITE)
    txn.put(b"key", b"value")
    db.close()
    with pytest.raises(AlopexError):
        txn.get(b"key")


def test_close_twice_is_error(db):
    db.close()
    with pytest.raises(AlopexError):
        db.close()



from typing import Any, Dict, List, Literal, Optional, Tuple, Union


class polars:
    class DataFrame:
        ...

    class LazyFrame:
        ...


class AlopexError(Exception):
    ...


class TxnMode:
    READ_ONLY: "TxnMode"
    READ_WRITE: "TxnMode"

    def __repr__(self) -> str: ...
    def __hash__(self) -> int: ...


class Metric:
    COSINE: "Metric"
    L2: "Metric"
    INNER_PRODUCT: "Metric"

    def __repr__(self) -> str: ...
    def __hash__(self) -> int: ...


class StorageMode:
    DISK: "StorageMode"
    IN_MEMORY: "StorageMode"

    def __repr__(self) -> str: ...
    def __hash__(self) -> int: ...


class HnswConfig:
    dim: int
    m: int
    ef_construction: int
    metric: Metric

    def __init__(
        self,
        dim: int,
        m: int = 16,
        ef_construction: int = 200,
        metric: Optional[Metric] = None,
    ) -> None: ...


class EmbeddedConfig:
    memory_limit_bytes: Optional[int]

    def __init__(self, memory_limit_bytes: Optional[int] = None) -> None: ...


class DatabaseOptions:
    path: Optional[str]
    storage_mode: StorageMode
    memory_limit_bytes: Optional[int]
    enable_metrics: bool

    def __init__(
        self,
        path: Optional[str] = None,
        storage_mode: Optional[StorageMode] = None,
        memory_limit_bytes: Optional[int] = None,
        enable_metrics: bool = False,
    ) -> None: ...


class SearchResult:
    key: bytes
    score: float
    metadata: Optional[bytes]
    vector: Optional[Any]  # numpy.ndarray[float32] when return_vectors=True

    def __init__(
        self,
        key: bytes,
        score: float,
        metadata: Optional[bytes] = None,
        vector: Optional[Any] = None,
    ) -> None: ...


class HnswStats:
    node_count: int
    deleted_count: int
    level_distribution: List[int]
    memory_bytes: int
    avg_edges_per_node: float

    def __init__(
        self,
        node_count: int = 0,
        deleted_count: int = 0,
        level_distribution: List[int] = [],
        memory_bytes: int = 0,
        avg_edges_per_node: float = 0.0,
    ) -> None: ...


class MemoryStats:
    total_bytes: int
    used_bytes: int
    free_bytes: int

    def __init__(self, total_bytes: int, used_bytes: int, free_bytes: int) -> None: ...

TableType = Literal[
    "MANAGED",
    "EXTERNAL",
    "VIEW",
    "MATERIALIZED_VIEW",
    "STREAMING_TABLE",
    "MANAGED_SHALLOW_CLONE",
    "FOREIGN",
    "EXTERNAL_SHALLOW_CLONE",
]

DataSourceFormat = Literal[
    "DELTA",
    "CSV",
    "JSON",
    "AVRO",
    "PARQUET",
    "ORC",
    "TEXT",
    "UNITY_CATALOG",
    "DELTASHARING",
    "DATABRICKS_FORMAT",
    "REDSHIFT_FORMAT",
    "SNOWFLAKE_FORMAT",
    "SQLDW_FORMAT",
    "SALESFORCE_FORMAT",
    "BIGQUERY_FORMAT",
    "NETSUITE_FORMAT",
    "WORKDAY_RAAS_FORMAT",
    "HIVE_SERDE",
    "HIVE_CUSTOM",
    "VECTOR_INDEX_FORMAT",
]

DeltaMode = Literal["error", "ignore", "append", "overwrite", "merge"]
CredentialProvider = Union[Literal["auto"], Dict[str, str]]


class CatalogInfo:
    name: str
    comment: Optional[str]
    storage_root: Optional[str]

    def __init__(
        self,
        name: str,
        comment: Optional[str] = None,
        storage_root: Optional[str] = None,
    ) -> None: ...


class NamespaceInfo:
    name: str
    catalog_name: str
    comment: Optional[str]
    storage_root: Optional[str]

    def __init__(
        self,
        name: str,
        catalog_name: str,
        comment: Optional[str] = None,
        storage_root: Optional[str] = None,
    ) -> None: ...


class ColumnInfo:
    name: str
    type_name: str
    position: int
    nullable: bool
    comment: Optional[str]

    def __init__(
        self,
        name: str,
        type_name: str,
        position: int = 0,
        nullable: bool = True,
        comment: Optional[str] = None,
    ) -> None: ...


class TableInfo:
    name: str
    catalog_name: str
    namespace_name: str
    table_type: TableType
    storage_location: Optional[str]
    data_source_format: Optional[DataSourceFormat]
    columns: List[ColumnInfo]
    primary_key: Optional[List[str]]
    comment: Optional[str]

    def __init__(
        self,
        name: str,
        catalog_name: str,
        namespace_name: str,
        table_type: TableType = "MANAGED",
        storage_location: Optional[str] = None,
        data_source_format: Optional[DataSourceFormat] = None,
        columns: List[ColumnInfo] = [],
        primary_key: Optional[List[str]] = None,
        comment: Optional[str] = None,
    ) -> None: ...


class Database:
    @staticmethod
    def open(path: str) -> "Database": ...

    @staticmethod
    def new() -> "Database": ...

    @staticmethod
    def open_in_memory() -> "Database": ...

    @staticmethod
    def open_with_config(config: EmbeddedConfig) -> "Database": ...

    def begin(self, mode: Optional[TxnMode] = None) -> "Transaction": ...
    def flush(self) -> None: ...
    def memory_usage(self) -> MemoryStats: ...
    def close(self) -> None: ...
    def create_hnsw_index(self, name: str, config: HnswConfig) -> None: ...
    def search_hnsw(
        self,
        name: str,
        query: Any,
        k: int,
        ef_search: Optional[int] = None,
    ) -> Tuple[List[SearchResult], HnswStats]: ...
    def drop_hnsw_index(self, name: str) -> None: ...
    def get_hnsw_stats(self, name: str) -> HnswStats: ...


class Transaction:
    def get(self, key: bytes) -> Optional[bytes]: ...
    def put(self, key: bytes, value: bytes) -> None: ...
    def delete(self, key: bytes) -> None: ...
    def upsert_vector(
        self,
        key: bytes,
        metadata: Optional[bytes],
        vector: Any,
        metric: Metric,
    ) -> None: ...
    def search_similar(
        self,
        query: Any,
        metric: Metric,
        k: int,
        filter_keys: Optional[List[bytes]] = None,
        return_vectors: bool = False,
        zero_copy_return: bool = True,
    ) -> List[SearchResult]: ...
    def get_vector(
        self,
        key: bytes,
        metric: Metric,
        zero_copy_return: bool = True,
    ) -> Any:
        """Get vector by key.

        Args:
            key: Vector key.
            metric: Metric (must match the metric used when storing).
            zero_copy_return: If True, uses zero-copy ownership transfer.

        Returns:
            numpy.ndarray[float32]: The vector data.

        Raises:
            KeyError: If the key does not exist.
        """
        ...
    def upsert_to_hnsw(
        self,
        name: str,
        key: bytes,
        vector: Any,
        metadata: Optional[bytes] = None,
    ) -> None: ...
    def delete_from_hnsw(self, name: str, key: bytes) -> None: ...
    def commit(self) -> None: ...
    def rollback(self) -> None: ...
    def __enter__(self) -> "Transaction": ...
    def __exit__(self, exc_type: Optional[Any], exc: Optional[Any], traceback: Optional[Any]) -> bool: ...


class Catalog:
    @staticmethod
    def list_catalogs() -> List[CatalogInfo]: ...

    @staticmethod
    def list_namespaces(catalog_name: str) -> List[NamespaceInfo]: ...

    @staticmethod
    def list_tables(catalog_name: str, namespace: str) -> List[TableInfo]: ...

    @staticmethod
    def get_table_info(
        catalog_name: str,
        namespace: str,
        table_name: str,
    ) -> TableInfo: ...

    @staticmethod
    def create_catalog(name: str) -> None: ...

    @staticmethod
    def delete_catalog(name: str) -> None: ...

    @staticmethod
    def create_namespace(catalog_name: str, namespace: str) -> None: ...

    @staticmethod
    def delete_namespace(catalog_name: str, namespace: str) -> None: ...

    @staticmethod
    def create_table(
        catalog_name: str,
        namespace: str,
        table_name: str,
        columns: List[ColumnInfo],
        storage_location: str,
        data_source_format: DataSourceFormat = "PARQUET",
    ) -> None: ...

    @staticmethod
    def delete_table(catalog_name: str, namespace: str, table_name: str) -> None: ...

    @staticmethod
    def scan_table(
        catalog_name: str,
        namespace: str,
        table_name: str,
        credential_provider: CredentialProvider = "auto",
        storage_options: Optional[Dict[str, str]] = None,
    ) -> "polars.LazyFrame": ...

    @staticmethod
    def write_table(
        df: "polars.DataFrame | polars.LazyFrame",
        catalog_name: str,
        namespace: str,
        table_name: str,
        delta_mode: DeltaMode = "error",
        storage_location: Optional[str] = None,
        credential_provider: CredentialProvider = "auto",
        storage_options: Optional[Dict[str, str]] = None,
        primary_key: Optional[List[str]] = None,
    ) -> None: ...



import importlib.util
import uuid

import pytest

from alopex import Database


def _module_available(name: str) -> bool:
    return importlib.util.find_spec(name) is not None


def _has_vector_api() -> bool:
    from alopex import Transaction

    return hasattr(Transaction, "upsert_vector")


def pytest_configure(config):
    config.addinivalue_line("markers", "requires_numpy: numpy が必要なテスト")
    config.addinivalue_line("markers", "requires_polars: polars が必要なテスト")


def pytest_runtest_setup(item):
    if "requires_numpy" in item.keywords:
        if not _module_available("numpy") or not _has_vector_api():
            pytest.skip("numpy feature が有効でないためスキップ")
    if "requires_polars" in item.keywords:
        if not _module_available("polars"):
            pytest.skip("polars が未インストールのためスキップ")


@pytest.fixture()
def db():
    db = Database.new()
    yield db
    try:
        db.close()
    except Exception:
        pass


@pytest.fixture()
def unique_name():
    return f"test_{uuid.uuid4().hex}"



import time

import pytest

pytest.importorskip("pytest_benchmark")
pl = pytest.importorskip("polars")

from alopex import Catalog, ColumnInfo

pytestmark = [
    pytest.mark.requires_polars,
    pytest.mark.requires_pytest_benchmark,
]


def _mean_time(fn, repeats=5):
    durations = []
    for _ in range(repeats):
        start = time.perf_counter()
        fn()
        durations.append(time.perf_counter() - start)
    return sum(durations) / len(durations)


def _prepare_catalog(tmp_path, unique_name, storage_location):
    catalog_name = f"{unique_name}_cat"
    namespace_name = f"{unique_name}_ns"
    table_name = f"{unique_name}_tbl"

    Catalog.create_catalog(catalog_name)
    Catalog.create_namespace(catalog_name, namespace_name)
    Catalog.create_table(
        catalog_name,
        namespace_name,
        table_name,
        [
            ColumnInfo("id", "INTEGER", 0, False),
            ColumnInfo("value", "DOUBLE", 1, False),
        ],
        storage_location,
    )
    return catalog_name, namespace_name, table_name


def _cleanup_catalog(catalog_name, namespace_name, table_name):
    try:
        Catalog.delete_table(catalog_name, namespace_name, table_name)
    except Exception:
        pass
    try:
        Catalog.delete_namespace(catalog_name, namespace_name)
    except Exception:
        pass
    try:
        Catalog.delete_catalog(catalog_name)
    except Exception:
        pass


def _make_large_df(rows):
    return pl.DataFrame(
        {
            "id": list(range(rows)),
            "value": [float(i) for i in range(rows)],
        }
    )


@pytest.mark.usefixtures("unique_name")
def test_scan_overhead_vs_polars(tmp_path, unique_name, benchmark):
    rows = 100_000
    df = _make_large_df(rows)
    storage_location = str(tmp_path / "data.parquet")
    df.write_parquet(storage_location)

    catalog_name, namespace_name, table_name = _prepare_catalog(
        tmp_path, unique_name, storage_location
    )

    def measure():
        direct = _mean_time(lambda: pl.scan_parquet(storage_location), repeats=10)
        wrapped = _mean_time(
            lambda: Catalog.scan_table(catalog_name, namespace_name, table_name),
            repeats=10,
        )
        overhead_ms = (wrapped - direct) * 1000.0
        return direct, wrapped, overhead_ms

    try:
        _, _, overhead_ms = benchmark.pedantic(
            measure, iterations=1, rounds=1, warmup_rounds=1
        )
        assert overhead_ms < 1.0, f"Python→Rust overhead too high: {overhead_ms:.3f}ms"
    finally:
        _cleanup_catalog(catalog_name, namespace_name, table_name)


@pytest.mark.usefixtures("unique_name")
def test_large_read_overhead(tmp_path, unique_name, benchmark):
    rows = 100_000
    df = _make_large_df(rows)
    storage_location = str(tmp_path / "data.parquet")
    df.write_parquet(storage_location)

    catalog_name, namespace_name, table_name = _prepare_catalog(
        tmp_path, unique_name, storage_location
    )

    def measure():
        direct = _mean_time(
            lambda: pl.scan_parquet(storage_location).collect(), repeats=5
        )
        wrapped = _mean_time(
            lambda: Catalog.scan_table(catalog_name, namespace_name, table_name).collect(),
            repeats=5,
        )
        overhead = (wrapped - direct) / direct
        return direct, wrapped, overhead

    try:
        _, _, overhead = benchmark.pedantic(
            measure, iterations=1, rounds=1, warmup_rounds=1
        )
        assert overhead < 0.05, f"read overhead too high: {overhead * 100:.2f}%"
    finally:
        _cleanup_catalog(catalog_name, namespace_name, table_name)


@pytest.mark.usefixtures("unique_name")
def test_large_write_overhead(tmp_path, unique_name, benchmark):
    rows = 100_000
    df = _make_large_df(rows)
    storage_location = str(tmp_path / "data.parquet")

    catalog_name = f"{unique_name}_cat"
    namespace_name = f"{unique_name}_ns"
    table_name = f"{unique_name}_tbl"

    Catalog.create_catalog(catalog_name)
    Catalog.create_namespace(catalog_name, namespace_name)

    try:
        Catalog.write_table(
            df,
            catalog_name,
            namespace_name,
            table_name,
            delta_mode="append",
            storage_location=storage_location,
        )

        def measure():
            direct = _mean_time(lambda: df.write_parquet(storage_location), repeats=5)
            wrapped = _mean_time(
                lambda: Catalog.write_table(
                    df,
                    catalog_name,
                    namespace_name,
                    table_name,
                    delta_mode="overwrite",
                ),
                repeats=5,
            )
            overhead = (wrapped - direct) / direct
            return direct, wrapped, overhead

        _, _, overhead = benchmark.pedantic(
            measure, iterations=1, rounds=1, warmup_rounds=1
        )
        assert overhead < 0.05, f"write overhead too high: {overhead * 100:.2f}%"
    finally:
        _cleanup_catalog(catalog_name, namespace_name, table_name)



import pytest

from alopex import Database, HnswConfig, TxnMode


@pytest.mark.requires_numpy
def test_hnsw_create_search_delete():
    import numpy as np

    db = Database.new()
    db.create_hnsw_index("idx", HnswConfig(2))
    with db.begin(TxnMode.READ_WRITE) as txn:
        vec = np.array([1.0, 0.0], dtype=np.float32)
        txn.upsert_to_hnsw("idx", b"k1", vec, None)
        txn.commit()

    results, stats = db.search_hnsw("idx", np.array([1.0, 0.0], dtype=np.float32), 1)
    assert len(results) >= 1
    assert stats.node_count >= 1

    with db.begin(TxnMode.READ_WRITE) as txn:
        txn.delete_from_hnsw("idx", b"k1")
        txn.commit()

    db.drop_hnsw_index("idx")


@pytest.mark.requires_numpy
def test_hnsw_multithreaded_search_releases_gil():
    import time
    import threading
    from concurrent.futures import ThreadPoolExecutor

    import numpy as np

    db = Database.new()
    name = "idx_mt"
    dim = 128
    count = 1000
    k = 10
    ef_search = 200

    db.create_hnsw_index(name, HnswConfig(dim))
    rng = np.random.default_rng(0)
    vectors = rng.standard_normal((count, dim)).astype(np.float32, copy=False)
    query = rng.standard_normal((dim,)).astype(np.float32, copy=False)

    with db.begin(TxnMode.READ_WRITE) as txn:
        for i in range(count):
            txn.upsert_to_hnsw(name, f"k{i}".encode(), vectors[i], None)
        txn.commit()

    threads = 4
    repeats = 10
    barrier = threading.Barrier(threads)

    def worker():
        barrier.wait()
        started = time.perf_counter()
        for _ in range(repeats):
            results, _stats = db.search_hnsw(name, query, k, ef_search=ef_search)
            assert len(results) == k
        finished = time.perf_counter()
        return started, finished

    with ThreadPoolExecutor(max_workers=threads) as pool:
        futures = [pool.submit(worker) for _ in range(threads)]
        start = time.perf_counter()
        main_ticks = 0
        while any(not f.done() for f in futures) and (time.perf_counter() - start) < 0.25:
            main_ticks += 1
        intervals = [future.result() for future in futures]

    overlap = False
    for i in range(len(intervals)):
        s1, e1 = intervals[i]
        for j in range(i + 1, len(intervals)):
            s2, e2 = intervals[j]
            if max(s1, s2) < min(e1, e2):
                overlap = True
                break
        if overlap:
            break

    assert (
        main_ticks > 1000
    ), "main thread did not make progress while workers were searching; GIL may not be released"
    assert overlap, "worker search_hnsw calls did not overlap; parallelism may be limited"



import threading

import pytest

pytest.importorskip("pytest_benchmark")
np = pytest.importorskip("numpy")

import alopex


pytestmark = [
    pytest.mark.requires_numpy,
    pytest.mark.requires_pytest_benchmark,
]


SCENARIOS = {
    "small": {"dim": 128, "count": 1_000, "k": 10},
    "medium": {"dim": 768, "count": 10_000, "k": 50},
    # NOTE: Keep "large" runnable on typical dev machines; original 100k can be too slow.
    "large": {"dim": 1536, "count": 5_000, "k": 100},
}


def _make_vectors(count: int, dim: int, seed: int):
    rng = np.random.default_rng(seed)
    vectors = rng.standard_normal((count, dim)).astype(np.float32, copy=False)
    query = rng.standard_normal((dim,)).astype(np.float32, copy=False)
    return vectors, query


def _populate_hnsw(db: "alopex.Database", index_name: str, vectors):
    # Reduce build-time cost for large runs while keeping behavior representative.
    cfg = alopex.HnswConfig(
        dim=int(vectors.shape[1]),
        m=16,
        ef_construction=50 if vectors.shape[0] >= 5_000 else 200,
        metric=alopex.Metric.COSINE,
    )
    db.create_hnsw_index(index_name, cfg)

    with db.begin(alopex.TxnMode.READ_WRITE) as txn:
        for i in range(vectors.shape[0]):
            txn.upsert_to_hnsw(index_name, key=f"hnsw_{i}".encode(), vector=vectors[i])
        txn.commit()


@pytest.fixture(scope="module")
def hnsw_env(request):
    scenario = request.param
    cfg = SCENARIOS[scenario]
    dim, count, k = cfg["dim"], cfg["count"], cfg["k"]

    vectors, query = _make_vectors(count=count, dim=dim, seed=2)
    db = alopex.Database.new()
    index_name = f"bench_{scenario}"
    _populate_hnsw(db, index_name, vectors)

    yield {
        "scenario": scenario,
        "dim": dim,
        "count": count,
        "k": k,
        "vectors": vectors,
        "query": query,
        "db": db,
        "index_name": index_name,
    }

    try:
        db.drop_hnsw_index(index_name)
    finally:
        db.close()


@pytest.mark.parametrize(
    "hnsw_env",
    [
        pytest.param("small", id="small"),
        pytest.param("medium", id="medium"),
        pytest.param("large", marks=pytest.mark.large, id="large"),
    ],
    indirect=True,
)
def test_bench_hnsw_search_latency(benchmark, hnsw_env):
    db = hnsw_env["db"]
    index_name = hnsw_env["index_name"]
    query = hnsw_env["query"]
    k = hnsw_env["k"]

    def run():
        _ = db.search_hnsw(index_name, query, k=k)

    iterations = 10 if hnsw_env["scenario"] != "large" else 1
    benchmark.pedantic(run, iterations=iterations, rounds=1, warmup_rounds=1)


@pytest.mark.parametrize(
    "hnsw_env",
    [
        pytest.param("small", id="small"),
        pytest.param("medium", id="medium"),
        pytest.param("large", marks=pytest.mark.large, id="large"),
    ],
    indirect=True,
)
def test_bench_hnsw_multithread_throughput_4_threads(benchmark, hnsw_env):
    db = hnsw_env["db"]
    index_name = hnsw_env["index_name"]
    query = hnsw_env["query"]
    k = hnsw_env["k"]

    threads = 4
    total_queries = 40 if hnsw_env["scenario"] != "large" else 8
    per_thread = total_queries // threads

    def worker():
        for _ in range(per_thread):
            _ = db.search_hnsw(index_name, query, k=k)

    def run():
        ts = [threading.Thread(target=worker) for _ in range(threads)]
        for t in ts:
            t.start()
        for t in ts:
            t.join()

    benchmark.pedantic(run, iterations=1, rounds=1, warmup_rounds=1)


@pytest.mark.parametrize(
    "hnsw_env",
    [
        pytest.param("small", id="small"),
        pytest.param("medium", id="medium"),
        pytest.param("large", marks=pytest.mark.large, id="large"),
    ],
    indirect=True,
)
def test_bench_hnsw_throughput_1_thread(benchmark, hnsw_env):
    db = hnsw_env["db"]
    index_name = hnsw_env["index_name"]
    query = hnsw_env["query"]
    k = hnsw_env["k"]

    total_queries = 40 if hnsw_env["scenario"] != "large" else 8

    def run():
        for _ in range(total_queries):
            _ = db.search_hnsw(index_name, query, k=k)

    benchmark.pedantic(run, iterations=1, rounds=1, warmup_rounds=1)



import pytest

from alopex import AlopexError, Catalog, ColumnInfo


def _cleanup_catalog(catalog_name, namespace_name=None, table_name=None):
    if table_name and namespace_name:
        try:
            Catalog.delete_table(catalog_name, namespace_name, table_name)
        except Exception:
            pass
    if namespace_name:
        try:
            Catalog.delete_namespace(catalog_name, namespace_name)
        except Exception:
            pass
    try:
        Catalog.delete_catalog(catalog_name)
    except Exception:
        pass


def test_metadata_api_fields(tmp_path, unique_name):
    catalog_name = f"{unique_name}_cat"
    namespace_name = f"{unique_name}_ns"
    table_name = f"{unique_name}_tbl"
    storage_location = str(tmp_path / "data.parquet")

    columns = [
        ColumnInfo("id", "INTEGER", 0, False),
        ColumnInfo("name", "TEXT", 1, True),
    ]

    Catalog.create_catalog(catalog_name)
    Catalog.create_namespace(catalog_name, namespace_name)
    Catalog.create_table(
        catalog_name,
        namespace_name,
        table_name,
        columns,
        storage_location,
        data_source_format="PARQUET",
    )

    try:
        catalogs = Catalog.list_catalogs()
        catalog_info = next(info for info in catalogs if info.name == catalog_name)
        assert catalog_info.comment is None
        assert catalog_info.storage_root is None

        namespaces = Catalog.list_namespaces(catalog_name)
        namespace_info = next(info for info in namespaces if info.name == namespace_name)
        assert namespace_info.catalog_name == catalog_name
        assert namespace_info.comment is None
        assert namespace_info.storage_root is None

        tables = Catalog.list_tables(catalog_name, namespace_name)
        table_info = next(info for info in tables if info.name == table_name)
        assert table_info.catalog_name == catalog_name
        assert table_info.namespace_name == namespace_name
        assert table_info.table_type == "MANAGED"
        assert table_info.storage_location == storage_location
        assert table_info.data_source_format == "PARQUET"
        assert table_info.primary_key is None
        assert table_info.comment is None
        assert len(table_info.columns) == 2
        assert table_info.columns[0].name == "id"
        assert table_info.columns[0].type_name == "INTEGER"
        assert table_info.columns[0].position == 0
        assert table_info.columns[0].nullable is False

        fetched = Catalog.get_table_info(catalog_name, namespace_name, table_name)
        assert fetched.name == table_name
        assert fetched.storage_location == storage_location
        assert fetched.data_source_format == "PARQUET"
        assert len(fetched.columns) == 2
    finally:
        _cleanup_catalog(catalog_name, namespace_name, table_name)


def test_ddl_error_cases(unique_name):
    catalog_name = f"{unique_name}_cat"
    namespace_name = f"{unique_name}_ns"
    table_name = f"{unique_name}_tbl"

    with pytest.raises(ValueError):
        Catalog.create_namespace(catalog_name, namespace_name)

    Catalog.create_catalog(catalog_name)
    try:
        with pytest.raises(RuntimeError):
            Catalog.create_catalog(catalog_name)

        columns = [ColumnInfo("id", "INTEGER", 0, False)]
        with pytest.raises(AlopexError):
            Catalog.create_table(
                catalog_name,
                namespace_name,
                table_name,
                columns,
                "/tmp/data.parquet",
                data_source_format="CSV",
            )
    finally:
        _cleanup_catalog(catalog_name, namespace_name, table_name)


@pytest.mark.requires_polars
def test_scan_table_storage_options_merge(tmp_path, unique_name, monkeypatch):
    import polars as pl

    catalog_name = f"{unique_name}_cat"
    namespace_name = f"{unique_name}_ns"
    table_name = f"{unique_name}_tbl"
    storage_location = str(tmp_path / "data.parquet")

    df = pl.DataFrame({"id": [1]})
    df.write_parquet(storage_location)

    Catalog.create_catalog(catalog_name)
    Catalog.create_namespace(catalog_name, namespace_name)
    Catalog.create_table(
        catalog_name,
        namespace_name,
        table_name,
        [ColumnInfo("id", "INTEGER", 0, False)],
        storage_location,
    )

    captured = {}

    def fake_scan(path, **kwargs):
        captured["path"] = path
        captured["kwargs"] = kwargs

        class DummyLazyFrame:
            def collect(self):
                return pl.DataFrame({"id": [1]})

        return DummyLazyFrame()

    monkeypatch.setattr(pl, "scan_parquet", fake_scan)

    try:
        Catalog.scan_table(
            catalog_name,
            namespace_name,
            table_name,
            credential_provider={"token": "old", "shared": "old"},
            storage_options={"shared": "new", "extra": "1"},
        )
        assert captured["path"] == storage_location
        assert captured["kwargs"]["token"] == "old"
        assert captured["kwargs"]["shared"] == "new"
        assert captured["kwargs"]["extra"] == "1"
    finally:
        _cleanup_catalog(catalog_name, namespace_name, table_name)


@pytest.mark.requires_polars
def test_scan_table_error_cases(unique_name):
    catalog_name = f"{unique_name}_cat"
    namespace_name = f"{unique_name}_ns"

    Catalog.create_catalog(catalog_name)
    Catalog.create_namespace(catalog_name, namespace_name)

    try:
        with pytest.raises(ValueError):
            Catalog.scan_table(catalog_name, namespace_name, "missing_table")
        with pytest.raises(ValueError):
            Catalog.scan_table("bad-name", namespace_name, "tbl")
    finally:
        _cleanup_catalog(catalog_name, namespace_name)


@pytest.mark.requires_polars
def test_write_table_modes(tmp_path, unique_name):
    import polars as pl

    catalog_name = f"{unique_name}_cat"
    namespace_name = f"{unique_name}_ns"
    table_name = f"{unique_name}_tbl"
    storage_location = str(tmp_path / "data.parquet")

    df_initial = pl.DataFrame({"id": [1, 2], "name": ["a", "b"]})
    df_append = pl.DataFrame({"id": [3], "name": ["c"]})
    df_overwrite = pl.DataFrame({"id": [10], "name": ["z"]})
    df_merge = pl.DataFrame({"id": [2, 4], "name": ["bb", "d"]})

    Catalog.create_catalog(catalog_name)
    Catalog.create_namespace(catalog_name, namespace_name)

    try:
        Catalog.write_table(
            df_initial,
            catalog_name,
            namespace_name,
            table_name,
            delta_mode="append",
            storage_location=storage_location,
        )
        out = pl.read_parquet(storage_location)
        assert out.shape[0] == 2

        Catalog.write_table(
            df_append,
            catalog_name,
            namespace_name,
            table_name,
            delta_mode="append",
        )
        out = pl.read_parquet(storage_location)
        assert out.shape[0] == 3

        Catalog.write_table(
            df_overwrite,
            catalog_name,
            namespace_name,
            table_name,
            delta_mode="overwrite",
        )
        out = pl.read_parquet(storage_location)
        assert out.shape[0] == 1
        assert out["id"].to_list() == [10]

        Catalog.write_table(
            df_merge,
            catalog_name,
            namespace_name,
            table_name,
            delta_mode="merge",
            primary_key=["id"],
        )
        out = pl.read_parquet(storage_location).sort("id")
        assert out["id"].to_list() == [2, 4, 10]
        assert out["name"].to_list() == ["bb", "d", "z"]
    finally:
        _cleanup_catalog(catalog_name, namespace_name, table_name)


@pytest.mark.requires_polars
def test_write_table_error_cases(tmp_path, unique_name):
    import polars as pl

    catalog_name = f"{unique_name}_cat"
    namespace_name = f"{unique_name}_ns"
    table_name = f"{unique_name}_tbl"
    storage_location = str(tmp_path / "data.parquet")
    df = pl.DataFrame({"id": [1]})

    Catalog.create_catalog(catalog_name)
    Catalog.create_namespace(catalog_name, namespace_name)

    try:
        with pytest.raises(AlopexError):
            Catalog.write_table(
                df,
                catalog_name,
                namespace_name,
                table_name,
                delta_mode="error",
                storage_location=storage_location,
            )

        with pytest.raises(AlopexError):
            Catalog.write_table(
                df,
                catalog_name,
                namespace_name,
                table_name,
                delta_mode="merge",
                storage_location=storage_location,
            )
    finally:
        _cleanup_catalog(catalog_name, namespace_name, table_name)



import pytest

pytest.importorskip("pytest_benchmark")

import alopex


def test_binding_overhead_get_put(benchmark):
    db = alopex.Database.new()
    key = b"12345678"
    value = b"abcdefgh"

    def run():
        txn = db.begin(alopex.TxnMode.READ_WRITE)
        txn.put(key, value)
        txn.commit()

        txn = db.begin()
        _ = txn.get(key)
        txn.rollback()

    benchmark.pedantic(
        run,
        iterations=10000,
        rounds=1,
        warmup_rounds=1,
    )



import importlib.util

import pytest

from alopex import AlopexError, Database, Metric, Transaction, TxnMode


def test_numpy_feature_flags():
    has_numpy = importlib.util.find_spec("numpy") is not None
    has_vector_api = hasattr(Transaction, "upsert_vector")

    if not has_vector_api:
        assert not hasattr(Transaction, "search_similar")
        assert not hasattr(Database, "search_hnsw")
        assert not hasattr(Transaction, "upsert_to_hnsw")
        assert not hasattr(Transaction, "delete_from_hnsw")
        assert not hasattr(Database, "create_hnsw_index")
        assert not hasattr(Database, "drop_hnsw_index")
        assert not hasattr(Database, "get_hnsw_stats")
        return

    if not has_numpy:
        db = Database.new()
        txn = db.begin(TxnMode.READ_WRITE)
        with pytest.raises(AlopexError):
            txn.upsert_vector(b"k1", None, [1.0], Metric.COSINE)
        return

    assert has_vector_api is True



import pytest

from alopex import Database, Metric, TxnMode


def _make_strided_f32(np, base):
    arr = np.asarray(base, dtype=np.float32)
    buf = np.zeros(arr.size * 2, dtype=np.float32)
    buf[::2] = arr
    return buf[::2]


@pytest.mark.requires_numpy
def test_upsert_vector_and_search_similar():
    import numpy as np

    db = Database.new()
    with db.begin(TxnMode.READ_WRITE) as txn:
        vec1 = np.array([1.0, 0.0, 0.0], dtype=np.float64)
        vec2 = np.array([0.0, 1.0, 0.0], dtype=np.float32)
        txn.upsert_vector(b"k1", None, vec1, Metric.COSINE)
        txn.upsert_vector(b"k2", b"meta", vec2, Metric.COSINE)
        results = txn.search_similar(vec1, Metric.COSINE, 2)

    assert len(results) >= 1
    keys = {result.key for result in results}
    assert b"k1" in keys


@pytest.mark.requires_numpy
@pytest.mark.parametrize(
    "query_factory",
    [
        pytest.param(
            lambda np, base: np.array(base, dtype=np.float32),
            id="f32_c_contig",
        ),
        pytest.param(
            lambda np, base: np.array(base, dtype=np.float64),
            id="f64_cast",
        ),
        pytest.param(
            lambda np, base: _make_strided_f32(np, base),
            id="f32_strided_view",
        ),
    ],
)
def test_search_similar_accepts_dtype_and_layout_variations(query_factory):
    import numpy as np

    db = Database.new()
    with db.begin(TxnMode.READ_WRITE) as txn:
        vec_a = np.array([1.0, 0.0, 0.0], dtype=np.float32)
        vec_b = np.array([0.0, 1.0, 0.0], dtype=np.float32)
        txn.upsert_vector(b"a", None, vec_a, Metric.COSINE)
        txn.upsert_vector(b"b", None, vec_b, Metric.COSINE)

        query = query_factory(np, [1.0, 0.0, 0.0])
        assert query.dtype in (np.float32, np.float64)
        results = txn.search_similar(query, Metric.COSINE, 2)

    keys = {r.key for r in results}
    assert b"a" in keys


@pytest.mark.requires_numpy
@pytest.mark.parametrize(
    "zero_copy_return",
    [pytest.param(True, id="zero_copy_return_true"), pytest.param(False, id="zero_copy_return_false")],
)
def test_search_similar_return_vectors_includes_ndarray(zero_copy_return):
    import numpy as np

    db = Database.new()
    with db.begin(TxnMode.READ_WRITE) as txn:
        vec = np.array([1.0, 2.0, 3.0], dtype=np.float32)
        txn.upsert_vector(b"k", None, vec, Metric.COSINE)
        results = txn.search_similar(
            vec.astype(np.float64),
            Metric.COSINE,
            1,
            return_vectors=True,
            zero_copy_return=zero_copy_return,
        )

    assert len(results) == 1
    assert results[0].key == b"k"
    assert results[0].vector is not None
    assert isinstance(results[0].vector, np.ndarray)
    assert results[0].vector.dtype == np.float32
    np.testing.assert_allclose(results[0].vector, vec, rtol=0, atol=0)


@pytest.mark.requires_numpy
@pytest.mark.parametrize(
    "vector_factory",
    [
        pytest.param(lambda np: np.array([4.0, 5.0, 6.0], dtype=np.float32), id="f32_c_contig"),
        pytest.param(lambda np: np.array([4.0, 5.0, 6.0], dtype=np.float64), id="f64_cast"),
        pytest.param(
            lambda np: _make_strided_f32(np, [4.0, 5.0, 6.0]),
            id="f32_strided_view",
        ),
    ],
)
def test_upsert_vector_accepts_dtype_and_layout_variations(vector_factory):
    import numpy as np

    db = Database.new()
    with db.begin(TxnMode.READ_WRITE) as txn:
        vec = vector_factory(np)
        txn.upsert_vector(b"k", None, vec, Metric.COSINE)
        out = txn.get_vector(b"k", Metric.COSINE, zero_copy_return=True)

    assert isinstance(out, np.ndarray)
    assert out.dtype == np.float32
    np.testing.assert_allclose(out, np.array([4.0, 5.0, 6.0], dtype=np.float32), rtol=0, atol=0)



import pytest

pytest.importorskip("pytest_benchmark")
np = pytest.importorskip("numpy")

import alopex

from .baseline_numpy import (
    cosine_similarity_loop,
    cosine_similarity_loop_over_get_vector,
    cosine_similarity_vectorized,
)


pytestmark = [
    pytest.mark.requires_numpy,
    pytest.mark.requires_pytest_benchmark,
]


SCENARIOS = {
    "small": {"dim": 128, "count": 1_000, "k": 10},
    "medium": {"dim": 768, "count": 10_000, "k": 50},
    # NOTE: Keep "large" runnable on typical dev machines; original 100k can be too slow.
    "large": {"dim": 1536, "count": 5_000, "k": 100},
}


def _make_vectors(count: int, dim: int, seed: int):
    rng = np.random.default_rng(seed)
    vectors = rng.standard_normal((count, dim)).astype(np.float32, copy=False)
    query = rng.standard_normal((dim,)).astype(np.float32, copy=False)
    return vectors, query


def _populate_vector_store(db: "alopex.Database", vectors):
    with db.begin(alopex.TxnMode.READ_WRITE) as txn:
        for i in range(vectors.shape[0]):
            txn.upsert_vector(
                key=f"vec_{i}".encode(),
                metadata=None,
                vector=vectors[i],
                metric=alopex.Metric.COSINE,
            )
        txn.commit()


@pytest.fixture(scope="module")
def vector_env(request):
    scenario = request.param
    cfg = SCENARIOS[scenario]
    dim, count, k = cfg["dim"], cfg["count"], cfg["k"]

    vectors, query = _make_vectors(count=count, dim=dim, seed=1)
    db = alopex.Database.new()
    _populate_vector_store(db, vectors)
    txn = db.begin()
    keys = [f"vec_{i}".encode() for i in range(count)]

    yield {
        "scenario": scenario,
        "dim": dim,
        "count": count,
        "k": k,
        "vectors": vectors,
        "query": query,
        "db": db,
        "txn": txn,
        "keys": keys,
    }

    try:
        txn.rollback()
    finally:
        db.close()


@pytest.fixture(scope="module")
def vector_env_conversion(request):
    scenario = request.param
    cfg = SCENARIOS[scenario]
    dim = cfg["dim"]

    # Conversion overhead should not be dominated by scan time, so keep the store small.
    vectors, query = _make_vectors(count=32, dim=dim, seed=11)
    db = alopex.Database.new()
    _populate_vector_store(db, vectors)
    txn = db.begin()

    yield {
        "scenario": scenario,
        "dim": dim,
        "k": 1,
        "query": query,
        "db": db,
        "txn": txn,
    }

    try:
        txn.rollback()
    finally:
        db.close()


@pytest.mark.parametrize(
    "vector_env_conversion",
    [
        pytest.param("small", id="small"),
        pytest.param("medium", id="medium"),
        pytest.param("large", marks=pytest.mark.large, id="large"),
    ],
    indirect=True,
)
@pytest.mark.parametrize(
    "query_kind",
    [
        pytest.param("f32_c", id="f32_c"),
        pytest.param("f64_c", id="f64_c"),
        pytest.param("strided", id="strided"),
    ],
)
def test_bench_vector_array_conversion_overhead(benchmark, vector_env_conversion, query_kind):
    dim, k = vector_env_conversion["dim"], vector_env_conversion["k"]
    txn = vector_env_conversion["txn"]

    if query_kind == "f32_c":
        query = vector_env_conversion["query"]
        iterations = 200
    elif query_kind == "f64_c":
        query = vector_env_conversion["query"].astype(np.float64, copy=True)
        iterations = 50
    elif query_kind == "strided":
        query = np.arange(dim * 2, dtype=np.float32)[::2]
        iterations = 50
    else:
        raise ValueError(f"unknown query_kind: {query_kind}")

    def run():
        _ = txn.search_similar(query, alopex.Metric.COSINE, k=k)

    benchmark.pedantic(
        run,
        iterations=iterations,
        rounds=1,
        warmup_rounds=1,
    )


@pytest.mark.parametrize(
    "vector_env",
    [
        pytest.param("small", id="small"),
        pytest.param("medium", id="medium"),
        pytest.param("large", marks=pytest.mark.large, id="large"),
    ],
    indirect=True,
)
def test_bench_vector_search_latency(benchmark, vector_env):
    k = vector_env["k"]
    query = vector_env["query"]
    txn = vector_env["txn"]

    def run():
        _ = txn.search_similar(query, alopex.Metric.COSINE, k=k)

    iterations = 10 if vector_env["scenario"] != "large" else 1
    benchmark.pedantic(run, iterations=iterations, rounds=1, warmup_rounds=1)


@pytest.mark.parametrize(
    "vector_env",
    [
        pytest.param("small", id="small"),
        pytest.param("medium", id="medium"),
        pytest.param("large", marks=pytest.mark.large, id="large"),
    ],
    indirect=True,
)
def test_bench_numpy_loop_baseline(benchmark, vector_env):
    def run():
        _ = cosine_similarity_loop(vector_env["query"], vector_env["vectors"])

    benchmark.pedantic(run, iterations=1, rounds=1, warmup_rounds=0)


@pytest.mark.parametrize(
    "vector_env",
    [
        pytest.param("small", id="small"),
        pytest.param("medium", id="medium"),
        pytest.param("large", marks=pytest.mark.large, id="large"),
    ],
    indirect=True,
)
def test_bench_numpy_vectorized_reference(benchmark, vector_env):
    def run():
        _ = cosine_similarity_vectorized(vector_env["query"], vector_env["vectors"])

    benchmark.pedantic(run, iterations=1, rounds=1, warmup_rounds=0)


@pytest.mark.parametrize(
    "vector_env",
    [
        pytest.param("small", id="small"),
        pytest.param("medium", id="medium"),
        pytest.param("large", marks=pytest.mark.large, id="large"),
    ],
    indirect=True,
)
def test_bench_numpy_loop_over_get_vector_baseline(benchmark, vector_env):
    keys = vector_env["keys"]
    query = vector_env["query"]
    txn = vector_env["txn"]

    def run():
        _ = cosine_similarity_loop_over_get_vector(txn, keys, query, alopex.Metric.COSINE)

    benchmark.pedantic(run, iterations=1, rounds=1, warmup_rounds=0)



import pytest

from alopex import AlopexError, Database, TxnMode


def test_put_get_delete_commit():
    db = Database.new()
    txn = db.begin(TxnMode.READ_WRITE)
    txn.put(b"key", b"value")
    assert txn.get(b"key") == b"value"
    txn.delete(b"key")
    txn.commit()

    txn2 = db.begin(TxnMode.READ_ONLY)
    assert txn2.get(b"key") is None


def test_read_only_put_is_error():
    db = Database.new()
    txn = db.begin(TxnMode.READ_ONLY)
    with pytest.raises(AlopexError):
        txn.put(b"key", b"value")


def test_commit_closes_transaction():
    db = Database.new()
    txn = db.begin(TxnMode.READ_WRITE)
    txn.commit()
    with pytest.raises(AlopexError):
        txn.get(b"key")


def test_context_manager_rolls_back():
    db = Database.new()
    with db.begin(TxnMode.READ_WRITE) as txn:
        txn.put(b"key", b"value")

    txn2 = db.begin(TxnMode.READ_ONLY)
    assert txn2.get(b"key") is None



import importlib.util

import pytest


def _module_available(name: str) -> bool:
    return importlib.util.find_spec(name) is not None


def pytest_configure(config):
    config.addinivalue_line("markers", "large: time-consuming benchmarks (excluded in CI)")
    config.addinivalue_line("markers", "requires_pytest_benchmark: pytest-benchmark が必要なテスト")


def pytest_runtest_setup(item):
    if "requires_pytest_benchmark" in item.keywords:
        if not _module_available("pytest_benchmark"):
            pytest.skip("pytest-benchmark が未インストールのためスキップ")

